/*-
 * Copyright (c) 2016 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <pthread.h>
#include <stdlib.h>

#include "vcl.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"
#include "cache/cache.h"
#include "vsb.h"
#include "miniobj.h"

#include "vcc_if.h"
#include "vre2/vre2.h"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod re2 error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod re2 error: " fmt, __VA_ARGS__)

struct vmod_re2_regex {
	unsigned	magic;
#define VMOD_RE2_REGEX_MAGIC 0x5c3f6f24
	pthread_key_t	ovk;
	vre2		*vre2;
};

#if 0
typedef struct ov_s {
	unsigned	magic;
#define OV_MAGIC 0x844bfa39
	const char	*subject;
	int		ovector[MAX_OV_USED];
} ov_t;
#endif

#if 0
static char c;
static const void *match_failed = (void *) &c;
#endif

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
        va_list args;

        va_start(args, fmt);
        if (ctx->method == VCL_MET_INIT) {
                AN(ctx->msg);
                VSB_vprintf(ctx->msg, fmt, args);
                VRT_handling(ctx, VCL_RET_FAIL);
        }
        else if (ctx->vsl)
                VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
        else
                /* Should this ever happen in vcl_fini() ... */
                VSL(SLT_VCL_Error, 0, fmt, args);
        va_end(args);
}

VCL_VOID
vmod_regex__init(const struct vrt_ctx *ctx, struct vmod_re2_regex **rep,
		 const char *vcl_name, VCL_STRING pattern)
{
	struct vmod_re2_regex *re;
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(rep);
	AZ(*rep);
	AN(vcl_name);
	AN(pattern);
	ALLOC_OBJ(re, VMOD_RE2_REGEX_MAGIC);
	AN(re);
	*rep = re;

	AZ(pthread_key_create(&re->ovk, NULL));
	err = vre2_init(&re->vre2, pattern);
	if (err != NULL) {
		vre2_fini(&re->vre2);
		re->vre2 = NULL;
		FREE_OBJ(re);
		*rep = NULL;
		VERR(ctx, "Cannot compile '%s' in regex object %s: %s", pattern,
		     vcl_name, err);
	}
}

VCL_VOID
vmod_regex__fini(struct vmod_re2_regex **rep)
{
	struct vmod_re2_regex *re;

	re = *rep;
	*rep = NULL;
	AZ(pthread_key_delete(re->ovk));
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	vre2_fini(&re->vre2);
	FREE_OBJ(re);
}

VCL_BOOL
vmod_regex_match(const struct vrt_ctx *ctx, struct vmod_re2_regex *re,
		 VCL_STRING subject)
{
	int match = 0;
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	if ((err = vre2_match(re->vre2, subject, &match)) != NULL) {
		VERR(ctx, ".match(): %s", err);
		return 0;
	}
	return match;
}

VCL_STRING
vmod_version(const struct vrt_ctx *ctx __attribute__((unused)))
{
	return VERSION;
}
